@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -23,6 +23,7 @@ from hypothesis.errors import NoExamples, NoSuchExample, Unsatisfiable, \
 from hypothesis.control import assume, reject
 from hypothesis.internal.compat import hrange
 from hypothesis.internal.reflection import get_pretty_function_description
+from hypothesis.internal.deferredformat import deferredformat
 
 
 def one_of_strategies(xs):
@@ -73,8 +74,6 @@ class SearchStrategy(object):
     hashable probably won't be.
 
     This may all seem overly complicated but it's for a fairly good reason.
-    For more discussion of the motivation see
-    http://hypothesis.readthedocs.org/en/master/internals.html
 
     Given these, data generation happens in three phases:
 
@@ -158,6 +157,10 @@ class SearchStrategy(object):
             strategy=self,
         )
 
+    @property
+    def branches(self):
+        return [self]
+
     def __or__(self, other):
         """Return a strategy which produces values by randomly drawing from one
         of this strategy or the other strategy.
@@ -191,9 +194,9 @@ class OneOfStrategy(SearchStrategy):
     """Implements a union of strategies. Given a number of strategies this
     generates values which could have come from any of them.
 
-    The conditional distribution draws uniformly at random from some non-empty
-    subset of these strategies and then draws from the conditional distribution
-    of that strategy.
+    The conditional distribution draws uniformly at random from some
+    non-empty subset of these strategies and then draws from the
+    conditional distribution of that strategy.
 
     """
 
@@ -228,6 +231,13 @@ class OneOfStrategy(SearchStrategy):
         for e in self.element_strategies:
             e.validate()
 
+    @property
+    def branches(self):
+        if self.bias is None:
+            return self.element_strategies
+        else:
+            return [self]
+
 
 class MappedSearchStrategy(SearchStrategy):
 
@@ -272,6 +282,13 @@ class MappedSearchStrategy(SearchStrategy):
                     raise
         reject()
 
+    @property
+    def branches(self):
+        return [
+            MappedSearchStrategy(pack=self.pack, strategy=strategy)
+            for strategy in self.mapped_strategy.branches
+        ]
+
 
 class FilteredStrategy(SearchStrategy):
 
@@ -293,14 +310,28 @@ class FilteredStrategy(SearchStrategy):
         self.filtered_strategy.validate()
 
     def do_draw(self, data):
-        for _ in hrange(3):
+        for i in hrange(3):
             start_index = data.index
             value = data.draw(self.filtered_strategy)
             if self.condition(value):
                 return value
             else:
+                if i == 0:
+                    data.note_event(deferredformat(
+                        'Retried draw from %r to satisfy filter', self,))
                 # This is to guard against the case where we consume no data.
                 # As long as we consume data, we'll eventually pass or raise.
                 # But if we don't this could be an infinite loop.
                 assume(data.index > start_index)
+        data.note_event('Aborted test because unable to satisfy %r' % (
+            self,
+        ))
         data.mark_invalid()
+
+    @property
+    def branches(self):
+        branches = [
+            FilteredStrategy(strategy=strategy, condition=self.condition)
+            for strategy in self.filtered_strategy.branches
+        ]
+        return branches
